/*********************************************************************************
* Projekt: Kontakt Alarm KA1																		*
* letzte nderung : 17.10.2007																	*
* Controller: ATmega88																				*
* Clock: interner Oszillator																		*	
*********************************************************************************/
#include <ioavr.h>
#include <inavr.h>

// Definitionen
#define NORMAL				0x00
#define VERZOEGERUNG		0x01
#define ALARM				0x02
#define LOW					0x01
#define HIGH				0x02
#define OFFEN				0x01
#define GESCHLOSSEN		0x00



#define KONTAKT (PINB & (1<<PB0))
#define Alarm_On   TCCR1A |= (1<<COM1A0); TCCR1B |= (1<<CS10); OCR1A=12;
#define Alarm_Off  TCCR1A &= ~(1<<COM1A0); TCCR1B &= ~(1<<CS10); PORTB &= ~(1<<PB1);

// globale Variablen
volatile unsigned char Alarmzeit = 10;
volatile unsigned char Alarmzaehler = 0;
volatile unsigned char Verzoegerungszeit = 5;
volatile unsigned char Verzoegerungszaehler = 0;
volatile unsigned char Batteriespannung = HIGH;
volatile unsigned char Ruhezustand = OFFEN;
unsigned char Mode = NORMAL;

/*********************************************************************************
*	Name:				init																				*
*	Beschreibung:	Initialisiert alle verwendeten Register								*
*  Parameter:		keine																				*
*  Rckgabewerte:	keine																				*
*********************************************************************************/
void init(void)
{
	// Port B
	PORTB = (1<<PB2) | (1<<PB3) | (1<<PB4) | (1<<PB5) | (1<<PB6) | (1<<PB7);
	DDRB = (1<<DDB1) ;		// PB1 als Ausgang konfigurieren
	
	// Port C
	PORTC = (1<<PC3) | (1<<PC4) | (1<<PC5) | (1<<PC6);	// Pull-Ups einschalten
	DDRC  = 0x00 ; 												// Port C als Eingang konfigurieren
	
	//Port D
	PORTD = 0xFF;										// Pull-Ups eingeschaltet
	DDRD = 0x00;										// PortD als Eingang konfigurieren
	
	// Timer 0 Initialisierung: CTC Mode
	TCCR0A = (1<<WGM01);
	TCCR0B = (1<<CS02) | (1 <<CS00);				// Prescaler 1024 -> 125 Hz Timer Clock
	TIMSK0 = (1<<OCIE0A);
	OCR0A = 125;										// 125 * 8 ms = 1 s
	
	// Timer 1 Initialisierung: Fast PWM-Mode zur Sounderzeugung 
	TCCR1A = (1<<WGM11) | (1<<WGM10); 
	TCCR1B = (1<<WGM13)  |(1<<WGM12);
	
	// Timer 2 Initialisierung:   CTC Mode
	TCCR2A  = (1<<WGM21);
	TCCR2B  = (1<<CS22)   | (1<<CS21);
	TIMSK2 = (1<<OCIE2A);
	OCR2A = 125;
	
	// AD-Wandler: Input Buffers fr die analogen Eingnge (auer PC3) abschalten
	DIDR0  = (1<<ADC0D) | (1<<ADC1D) | (1<<ADC2D) | (1<<ADC4D) | (1<<ADC5D);    
	
	// Pinchange Interrupt fr PCINT0 aktivieren
	PCMSK0 = (1<<PCINT0);  	
	PCICR  = (1<<PCIE0);
	
	// Power Reduction Register: TWI, USART und SPI Schnittstelle deaktivieren
	PRR = (1<<PRTWI) | (1<<PRUSART0) | (1<<PRSPI);
}

/*********************************************************************************
*								Timer 0 Compare Match Interrupt									*
*	Beschreibung:	Wird im Sekundentakt aufgerufen und realisiert die Verzge-		*
*						rungs und Alarmzeiten														*
*  Parameter:		keine																				*
*  Rckgabewerte:	keine																				*
*********************************************************************************/
#pragma vector=TIMER0_COMPA_vect
__interrupt void sekundentakt(void)
{
	unsigned char zustand = Ruhezustand;	//	Zustand des Jumpers sichern
	
	if(Mode == VERZOEGERUNG)					// ist der Verzgerungsmodus aktiv?
	{
		if( KONTAKT != zustand)					// wird der Kontakt noch bettigt?
		{
			if(Verzoegerungszaehler)			// ist die Verzgerungszeit abgelaufen?
				Verzoegerungszaehler--;
			else										// ja, dann in den Alarmmodus wechseln 
			{											//	und den Alarm einschalten	
				Mode = ALARM;
				Alarmzaehler = Alarmzeit;
				Alarm_On
			}
		}
		else											// nein, der Kontakt wird nicht mehr bettigt
			Mode = NORMAL;
	}
	else if(Mode == ALARM)						// ist der Alarmmodus aktiv?
	{
		if(KONTAKT != zustand)					// wird der Kontakt noch bettigt?
		{
			if(Alarmzaehler)						// ist die Alarmzeit abgelaufen?
				Alarmzaehler--;
			else										// ja, dann in den Normalmodus wechseln
			{											//	und den Alarm ausschalten	
				Mode = NORMAL;
				Alarm_Off
			}
		}
		else											// nein, der Kontakt wird nicht mehr bettigt
		{
			Mode = NORMAL;
			Alarm_Off
		}
	}
}

/*********************************************************************************
*								Timer 2 Compare Match Interrupt									*
*	Beschreibung:	Wird alle 250 ms aufgerufen und wechselt die Alarmfrequenz.		*
*						Ist die Batteriespannung zu niedrig, wird der Alarm an oder 	*
*						ausgeschaltet.																	*
*  Parameter:		keine																				*
*  Rckgabewerte:	keine																				*
*********************************************************************************/
#pragma vector=TIMER2_COMPA_vect
__interrupt void TimeBase(void)
{
	static unsigned char toggle = 0;

	if( Mode == ALARM )								// ist der Alarmmodus aktiv?
	{
			if(toggle == 0)
			{
				OCR1A = 24;
				if(Batteriespannung == LOW)		// ist die Batteriespannung zu niedrig?
				{
					Alarm_On								// muss in Klammern stehen, da es 
				}											// drei Anweisungen ersetzt wird
				toggle = 1;
			}
			else
			{
				if(Batteriespannung == HIGH)		// ist die Batteriespannung ok?
					OCR1A = 12;
				else
				{
					Alarm_Off							// muss in Klammern stehen, da es 
				}											// drei Anweisungen ersetzt wird
				toggle = 0;
			}
	}
}

/*********************************************************************************
*								Pin Change Interrupt													*
*	Beschreibung:	Wird aufgerufen sobald ein Pegelwechsel an Pin PB0 erkannt		*
*						wird. Der Interrupt beendet den Power Down Mode und setzt das 	*
*						Sleep Enable Bit zurck. Hat der Pin den richtigen Pegel, wird	*
*						der Verzgerungsmodus aktiviert und der Interrupt abgeschaltet.*
*  Parameter:		keine																				*
*  Rckgabewerte:	keine																				*
*********************************************************************************/
#pragma vector=PCINT0_vect
__interrupt void pinchange(void)
{  
	unsigned char zustand = Ruhezustand;
	
	SMCR &= ~(1<<SE);         	// Sleep Enable Bit zurcksetzen

	if( (KONTAKT != zustand) && (Mode == NORMAL) )
	{
		Mode = VERZOEGERUNG;
		Verzoegerungszaehler = Verzoegerungszeit;
		PCICR  = 0x00;  			// PinChange Interrupt ausschalten
	}
}


/*********************************************************************************
*	Name:				Timer 2 Compare Match Interrupt											*
*	Beschreibung:	Startet die AD-Wandlung des analogen Signals an ADC0 oder 		*
*						ADC1 oder ADC2. Das Ergebnis wird zurckgegeben.					*
*  Parameter:		Kanal: 1 bis 3, entpricht (ADC0 bis ADC2)								*
*  Rckgabewerte:	Digitalisierter Analogwert mit 8 Bit Auflsung						*
*********************************************************************************/
unsigned char get_adc_value(char kanal)
{
	// AD-Wandler einschalten, Prescaler 2, Interrupt Flag lschen
	ADCSRA = (1<<ADEN) | (1<<ADPS0) | (1<<ADIF);     
	ADMUX = (1<<REFS1) | (1<<REFS0) | (1<<ADLAR) | (kanal-1) ;  // Ref intern 1,1V
	__delay_cycles(1920);					// Warten bis der Kondensator aufgeladen ist
	ADCSRA |= (1<<ADSC);          		// AD-Wandlung starten
	while (!(ADCSRA & (1<<ADIF)));     	// Warten bis Messung fertig
	ADCSRA = 0;                   		// Ad-Wandler abschalten
	return ADCH;                        // Rckgabewert 8bit
}

/*********************************************************************************
*	Name:				berechne_zeit																	*
*	Beschreibung:	Rechnet die digitalisierten Werte in Sekunden um. 					*
*  Parameter:		Digitalisierter Analogwert (0-255), Kanal (0,1)						*
*  Rckgabewerte:	Berechnete Sekunden															*
*********************************************************************************/
unsigned char berechne_zeit(unsigned char wert, unsigned char kanal)
{
   unsigned int temp_wert;
	
	if(kanal == 0)
		temp_wert = (wert * 100) / 423;
	
	else if(kanal == 1)
		temp_wert = (wert * 100) / 212;
	
	return temp_wert;
}

/*********************************************************************************
*	Name:				main																				*
*	Beschreibung:	Haupfunktion mit Endlosschleife, die sich zyklisch wiederholt 	*					
*  Parameter:		keine																				*
*  Rckgabewerte:	keine																				*
*********************************************************************************/

void main( void )
{
	static unsigned char adc_abfrage = 0;
	
	init();										// Funktion initialisiert alle notwendigen Register
	__enable_interrupt();      			// Interrupts zulassen
	while(1)										// Hauptschleife
	{
		if(adc_abfrage == 0)					// wurden die Potis und die Batteriespannung bereits 
		{											// abgefragt?
			Alarmzeit = berechne_zeit(get_adc_value(1),0);			// Poti R3 abfragen, Zeit berechnen
			Verzoegerungszeit = berechne_zeit(get_adc_value(2),1);// Poti R2 abfragen, Zeit berechnen
			
			if(get_adc_value(3) < 145)		// ist die Batteriespannung kleiner 2 V (145)
				Batteriespannung = LOW;											
			
			if( PINC & (1<<PC3) )			//	Jumper J1 abfragen und Ruhezustand ermitteln									
				Ruhezustand = OFFEN;
			else
				Ruhezustand = GESCHLOSSEN;
			
			adc_abfrage = 1;						
		}
		
		if(Mode == NORMAL)					// wenn der Kontakt nicht ausgelst wurde geht der 
		{											// Controller in den Power Down Mode
			adc_abfrage = 0;
			Batteriespannung = HIGH;
			PCIFR = (1<<PCIF0);
			PCICR  = (1<<PCIE0);  			// PinChange Interrupt fr PBO wieder einschalten
			SMCR = (1<<SM1) | (1<<SE);		// Power-down Mode
		 	__sleep();
			
		 	SMCR &= ~(1<<SE);         		// Sleep Enable Bit zurcksetzen
		}
	}
}
